import os
import requests
import json
import datetime
import pytz
from . import ServiceInterface, CalendarInterface, EventSerializer, BaseEvent, POTATOTIME_EVENT_SUBJECT, POTATOTIME_EVENT_DESCRIPTION
from potatotime.storage import Storage, FileStorage
from typing import Optional, List, Dict
from msal import ConfidentialClientApplication, SerializableTokenCache
from .auth import get_auth_code


class _MicrosoftEventSerializer(EventSerializer):
    def serialize(self, field_name: str, event: BaseEvent):
        if field_name in ('start', 'end'):
            field_value = getattr(event, field_name)
            return field_name, {
                'dateTime': field_value.isoformat(),
                'timeZone': getattr(field_value.tzinfo, 'zone', field_value.tzname()) if field_value.tzinfo else 'UTC'
            }
        if field_name == 'is_all_day':
            return 'isAllDay', event.is_all_day
        raise NotImplementedError(f"Serializing {field_name} is not supported")
    
    def deserialize(self, field_name: str, event_data: dict):
        if field_name == 'id':
            return event_data.get('id')
        if field_name in ('start', 'end'):
            time = datetime.datetime.fromisoformat(event_data[field_name]['dateTime'])
            timezone = pytz.timezone(event_data[field_name]['timeZone'])
            time = timezone.localize(time)
            return time
        if field_name == 'url':
            return event_data.get('webLink')
        if field_name == 'source_event_id':
            return event_data.get('singleValueExtendedProperties', [{}])[0].get('value')
        if field_name == 'declined':
            # TODO: Not fully implemented. Pass the user's email address to finish implementing
            return any([
                attendee['status']['response'] == 'declined' and attendee['emailAddress']['address'] == None
                for attendee in event_data.get('attendees', [])
            ])
        if field_name == 'is_all_day':
            return event_data.get('isAllDay', False)


class MicrosoftService(ServiceInterface):

    def __init__(self):
        self.client_id = os.environ['POTATOTIME_MSFT_CLIENT_ID']
        self.client_secret = os.environ['POTATOTIME_MSFT_CLIENT_SECRET']
        self.redirect_uri = 'http://localhost:8080'
        self.scopes = ['Calendars.ReadWrite']
        self.event_serializer = _MicrosoftEventSerializer()

        self.cache = SerializableTokenCache()
        self.app = ConfidentialClientApplication(
            self.client_id,
            authority='https://login.microsoftonline.com/common',
            client_credential=self.client_secret,
            token_cache=self.cache,
        )
        self.access_token = None

    def authorize(self, user_id: str, storage: Storage=FileStorage(), interactive: bool=True):
        accounts = []
        if storage.has_user_credentials(user_id):
            # NOTE: Technically, this one cache should be used to globally
            # contain all accounts. 'Sharding' cache across files in case there
            # are many accounts.
            self.cache.deserialize(storage.get_user_credentials(user_id))
            accounts = self.app.get_accounts()
            
        if accounts:
            account = accounts[0]  # assumes only one account per
            result = self.app.acquire_token_silent(self.scopes, account)  # handles refreshing access token
            self.access_token = result['access_token']
            storage.save_user_credentials(user_id, self.cache.serialize())

        if not self.access_token and interactive:
            auth_url = self.app.get_authorization_request_url(self.scopes, redirect_uri=self.redirect_uri)
            auth_code = get_auth_code(auth_url)
            token_response = self.app.acquire_token_by_authorization_code(
                auth_code,
                scopes=self.scopes,
                redirect_uri=self.redirect_uri
            )
            self.access_token = token_response['access_token']
            storage.save_user_credentials(user_id, self.cache.serialize())
        
        if not self.access_token:
            raise Exception('No credentials found, or credentials are expired.')
    
    def list_calendars(self) -> List[Dict]:
        url = "https://graph.microsoft.com/v1.0/me/calendars"
        headers = {
            'Authorization': f'Bearer {self.access_token}',
            'Content-Type': 'application/json'
        }
        response = requests.get(url, headers=headers)
        response.raise_for_status()
        calendar_list = response.json()
        return calendar_list.get('value', [])
    
    # TODO: duplicated from GoogleCalendar
    def get_calendar(self, calendar_id: Optional[str]=None):
        calendars = self.list_calendars()
        for calendar in calendars:
            if calendar['id'] == calendar_id or calendar_id is None:
                return MicrosoftCalendar(self, calendar)
        raise ValueError(f'Invalid calendar_id: {calendar_id}')
    

class MicrosoftCalendar(CalendarInterface):
    def __init__(self, service, calendar_id):
        self.service = service
        self.calendar_id = calendar_id
        self.event_serializer = _MicrosoftEventSerializer()
    
    def get_events(
        self,
        start: Optional[datetime.datetime]=None,
        end: Optional[datetime.datetime]=None,
        max_events: int=1000,
        results_per_page: int=100,
    ):
        url = 'https://graph.microsoft.com/v1.0/me/calendarView'
        headers = {
            'Authorization': f'Bearer {self.service.access_token}'
        }
        
        if not start:
            start = datetime.datetime.utcnow()
        if not end:
            end = start + datetime.timedelta(days=30)
        
        params = {
            'startDateTime': start.isoformat() + 'Z',
            'endDateTime': end.isoformat() + 'Z',
            '$orderby': 'start/dateTime',
            '$top': results_per_page,
            '$expand': "singleValueExtendedProperties($filter=id eq 'String {66f5a359-4659-4830-9070-00040ec6ac6e} Name potatotime')",
        }
        
        events = []
        next_link = None

        while True:
            if next_link:
                response = requests.get(next_link, headers=headers)
            else:
                response = requests.get(url, headers=headers, params=params)

            response.raise_for_status()
            response_data = response.json()
            events.extend(response_data.get('value', []))
            
            # Check if we have reached the maximum number of events
            if len(events) >= max_events:
                events = events[:max_events]
                break
            
            # Check if there's a next page
            next_link = response_data.get('@odata.nextLink')
            if not next_link:
                break
        
        return events

    def create_event(self, event_data: dict, source_event_id: Optional[str]):
        url = 'https://graph.microsoft.com/v1.0/me/events'
        headers = {
            'Authorization': f'Bearer {self.service.access_token}',
            'Content-Type': 'application/json'
        }
        event_data['subject'] = POTATOTIME_EVENT_SUBJECT
        event_data['body'] = {
            "contentType": "HTML",
            "content": POTATOTIME_EVENT_DESCRIPTION
        }
        event_data["singleValueExtendedProperties"] = [{
            "id": "String {66f5a359-4659-4830-9070-00040ec6ac6e} Name potatotime",
            "value": source_event_id,
        }]
        response = requests.post(url, headers=headers, json=event_data)
        response.raise_for_status()
        event = response.json()
        print(f"Event created: {event['webLink']}")
        return event

    def update_event(self, event_id, update_data):
        # TODO: check the event is potatotime-created
        url = f'https://graph.microsoft.com/v1.0/me/events/{event_id}'
        headers = {
            'Authorization': f'Bearer {self.service.access_token}',
            'Content-Type': 'application/json'
        }
        response = requests.patch(url, headers=headers, json=update_data)
        response.raise_for_status()
        event = response.json()
        print(f"Event updated: {event['webLink']}")
        return event

    def delete_event(self, event_id):
        # TODO: check the event is potatotime-created
        url = f'https://graph.microsoft.com/v1.0/me/events/{event_id}'
        headers = {
            'Authorization': f'Bearer {self.service.access_token}'
        }
        response = requests.delete(url, headers=headers)
        response.raise_for_status()
        print(f'Event "{event_id}" deleted.')
        return response.status_code



from abc import ABC, abstractmethod
from dataclasses import dataclass, fields
from datetime import datetime
from typing import Optional
from potatotime.storage import Storage, FileStorage
import pytz


# TODO: new home for constants?
POTATOTIME_EVENT_SUBJECT = "Busy"
POTATOTIME_EVENT_DESCRIPTION = "Synchronized by PotatoTime 🥔"


class ServiceInterface(ABC):
    @abstractmethod
    def authorize(self, user_id: str, storage: Storage=FileStorage(), interactive: bool=True):
        """

        :param interactive: If true, will prompt user for authentication via a
            browser automatically. If not, throw an error when authentication
            fails.
        """
        pass


class CalendarInterface(ABC):
    event_serializer: 'EventSerializer'

    @abstractmethod
    def get_events(
        self,
        start: Optional[datetime]=None,
        end: Optional[datetime]=None,
        max_events: int=1000,
    ):
        pass

    @abstractmethod
    def create_event(self, event_data):
        pass

    @abstractmethod
    def update_event(self, event_id, update_data):
        pass

    @abstractmethod
    def delete_event(self, event_id):
        pass


class EventSerializer(ABC):
    @abstractmethod
    def serialize(self, field_name: str):
        pass

    @staticmethod
    @abstractmethod
    def deserialize(self, field_name: str, data: dict):
        pass


@dataclass
class BaseEvent:
    start: datetime
    end: datetime
    is_all_day: bool

    @classmethod
    def from_(cls, other: 'BaseEvent'):
        return cls(**{
            field.name: getattr(other, field.name)
            for field in fields(cls)
        })
    
    def __eq__(self, other: 'BaseEvent'):
        for field in fields(self):
            self_value = getattr(self, field.name)
            other_value = getattr(other, field.name)
            if field.type is datetime:
                self_value = self_value.astimezone(pytz.utc)
                other_value = other_value.astimezone(pytz.utc)
            if self_value != other_value:
                return False
        return True


@dataclass
class StubEvent(BaseEvent):
    """Used to serialize payloads for APIs"""
    def serialize(self, serializer: EventSerializer) -> dict:
        payload = {}
        for field in fields(self):
            key, value = serializer.serialize(field.name, self)
            if key is not None:
                payload[key] = value
        return payload


@dataclass
class CreatedEvent(BaseEvent):
    """Used to standardize event payloads returned by APIs"""
    id: str
    url: str
    
    @classmethod
    def deserialize(cls, event_data: dict, serializer: EventSerializer):
        return cls(**{
            field.name: serializer.deserialize(field.name, event_data)
            for field in fields(cls)
        })


@dataclass
class ExtendedEvent(CreatedEvent):
    """Used to extract additional information from payloads returned by APIs"""
    declined: bool = False
    source_event_id: Optional[str] = None



from typing import List
from .services import CalendarInterface, ExtendedEvent, StubEvent
import datetime


def synchronize(
    calendars: List[CalendarInterface],
    max_days: int=365,
    max_events: int=1000
):
    start = datetime.datetime.utcnow()
    end = start + datetime.timedelta(days=max_days)
    calendars_events = [
        [
            ExtendedEvent.deserialize(event_data, calendar.event_serializer)
            for event_data in calendar.get_events(start=start, end=end, max_events=max_events)
        ]
        for calendar in calendars
    ]

    created, updated, deleted = {}, {}, {}
    for i in range(len(calendars)):
        for j in range(len(calendars)):
            if i == j:
                continue
            created[(i, j)], updated[(i, j)], deleted[(i, j)] = \
                synchronize_from_to(calendars[i], calendars_events[i], calendars[j], calendars_events[j])
    return created, updated, deleted


def synchronize_from_to(
    calendar1: CalendarInterface,
    events1: List[ExtendedEvent],
    calendar2: CalendarInterface,
    events2: List[ExtendedEvent]
) -> List[str]:
    created, updated, deleted = [], [], []

    # Delete any duplicated event copies
    source_event_ids = {}
    for event in events2:
        if event.source_event_id:
            if event.source_event_id in source_event_ids:
                deleted.append(event)
                calendar2.delete_event(event.id)
            source_event_ids[event.source_event_id] = event

    for event1 in events1:
        # Handle edited events
        if event1.id in source_event_ids:  # events already sync'ed
            event2 = source_event_ids.pop(event1.id)
            copy_stub = StubEvent.from_(event2)
            orig_stub = StubEvent.from_(event1)
            if copy_stub == orig_stub:  # if still equal to original, we're done
                continue

            copy_data = orig_stub.serialize(calendar2.event_serializer)
            copy_data = calendar2.update_event(event2.id, copy_data)
            copy = ExtendedEvent.deserialize(copy_data, calendar2.event_serializer)
            updated.append(copy)
            continue

        # Handle newly-created events
        if (  # Do not copy any of the following events
            event1.source_event_id is not None  # copy created by PotatoTime
            or event1.declined  # event declined by user (only implemented for Google)
        ):
            continue

        copy2_data = StubEvent.from_(event1).serialize(calendar2.event_serializer)
        copy2_data = calendar2.create_event(copy2_data, source_event_id=event1.id)
        copy2 = ExtendedEvent.deserialize(copy2_data, calendar2.event_serializer)
        created.append(copy2)

    # Delete any remaining events (e.g., events that were deleted in the source calendar)
    for event in source_event_ids.values():
        deleted.append(event)
        calendar2.delete_event(event.id)

    return created, updated, deleted


import caldav
import datetime
import os
from typing import Optional, List, Dict
from . import ServiceInterface, CalendarInterface, EventSerializer, BaseEvent, POTATOTIME_EVENT_SUBJECT, POTATOTIME_EVENT_DESCRIPTION
from potatotime.storage import Storage, FileStorage


class _AppleEventSerializer(EventSerializer):
    def serialize(self, field_name: str, event: BaseEvent):
        if field_name in ('is_all_day',):
            return None, None # TODO: implement me
        if field_name in ('start', 'end'):
            return field_name, getattr(event, field_name)
        raise NotImplementedError(f"Serializing {field_name} is not supported")
    
    def deserialize(self, field_name: str, event_data):
        if field_name == 'id':
            return event_data.instance.vevent.uid.value
        if field_name in ('start', 'end'):
            return getattr(event_data.instance.vevent, f"dt{field_name}").value
        if field_name in ('url', 'source_event_id', 'declined'):
            return None  # TODO: implement me


class AppleService(ServiceInterface):
    
    def __init__(self):
        self.client = caldav.DAVClient(
            url='https://caldav.icloud.com/',
            username=os.environ['POTATOTIME_APPLE_USERNAME'],
            password=os.environ['POTATOTIME_APPLE_PASSWORD'],
        )
        self.principal = self.client.principal()
        self.calendars = self.principal.calendars()
        self.event_serializer = _AppleEventSerializer()

    def authorize(self, user_id: str, storage: Storage=FileStorage(), interactive: bool=True):
        # Authorization is handled in the constructor for Apple Calendar
        pass

    def list_calendars(self) -> List[Dict]:
        return [{'id': cal.url, 'name': cal.name, 'object': cal} for cal in self.calendars]

    # TODO: duplicated from GoogleCalendar
    def get_calendar(self, calendar_id: Optional[str]=None):
        calendars = self.list_calendars()
        for calendar in calendars:
            if calendar['id'] == calendar_id or calendar_id is None:
                return AppleCalendar(self, calendar['object'])
        raise ValueError(f'Invalid calendar_id: {calendar_id}')


class AppleCalendar(CalendarInterface):

    def __init__(self, service, calendar):
        self.service = service
        self.calendar = calendar
        self.event_serializer = _AppleEventSerializer()

    def create_event(self, event_data):
        event = f"""
BEGIN:VCALENDAR
VERSION:2.0
BEGIN:VEVENT
UID:{datetime.datetime.now().timestamp()}@example.com
DTSTART:{event_data['start'].strftime('%Y%m%dT%H%M%S')}
DTEND:{event_data['end'].strftime('%Y%m%dT%H%M%S')}
SUMMARY:{POTATOTIME_EVENT_SUBJECT}
DESCRIPTION:{POTATOTIME_EVENT_DESCRIPTION}
LOCATION:{event_data.get('location', '')}
END:VEVENT
END:VCALENDAR
"""
        new_event = self.calendar.add_event(event)
        print(f"Event '{new_event.instance.vevent.uid.value}' created with UID: {new_event.instance.vevent.uid.value}")
        return new_event

    def update_event(self, event, update_data):
        if event:
            component = event.vobject_instance.vevent
            if 'start' in update_data:
                component.dtstart.value = update_data['start']
            if 'end' in update_data:
                component.dtend.value = update_data['end']
            if 'summary' in update_data:
                component.summary.value = update_data['summary']
            if 'description' in update_data:
                component.description.value = update_data['description']
            if 'location' in update_data:
                component.location.value = update_data['location']
            event.save()
            print(f"Event '{event.vobject_instance.vevent.uid.value}' edited.")

    def delete_event(self, event):
        event.delete()
        print(f'Event "{event.instance.vevent.uid.value}" deleted')

    def get_events(
        self,
        start: Optional[datetime.datetime]=None,
        end: Optional[datetime.datetime]=None,
        max_events: int=1000,
    ):  
        if not start:
            start = datetime.datetime.utcnow()
        if not end:
            end = start + datetime.timedelta(days=30)
        
        all_events = []
        events = self.calendar.date_search(start=start, end=end)
        
        for event in events:
            all_events.append(event)
            if len(all_events) >= max_events:
                return all_events[:max_events]

        return all_events


import webbrowser
from urllib.parse import urlparse, parse_qs
from http.server import BaseHTTPRequestHandler, HTTPServer


class OAuthHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        parsed_url = urlparse(self.path)
        query_params = parse_qs(parsed_url.query)
        if 'code' in query_params:
            self.server.auth_code = query_params['code'][0]
            self.send_response(200)
            self.send_header('Content-type', 'text/html')
            self.end_headers()
            self.wfile.write(b'You can close this window now.')
        else:
            self.send_response(400)
            self.send_header('Content-type', 'text/html')
            self.end_headers()
            self.wfile.write(b'Authorization failed.')


def get_auth_code(url: str, port: int=8080):
    webbrowser.open(url)
    httpd = HTTPServer(('localhost', port), OAuthHandler)
    httpd.handle_request()
    return httpd.auth_code


from potatotime.services import StubEvent, CreatedEvent
from potatotime.services.gcal import GoogleService
from potatotime.services.outlook import MicrosoftService
from potatotime.synchronize import synchronize
from utils import TIMEZONE, TEST_GOOGLE_USER_ID, TEST_MICROSOFT_USER_ID
import datetime
import pytz


NOW = datetime.datetime.now().replace(second=0, microsecond=0)
TMW_START = NOW + datetime.timedelta(days=1)
TMW_END = NOW + datetime.timedelta(days=1, hours=1)
DAT_START = NOW + datetime.timedelta(days=2) # "Day After Tomorrow" xD
DAT_END = NOW + datetime.timedelta(days=2, hours=1)
TMW_DAY = TMW_START.replace(hour=0, minute=0, second=0, microsecond=0)
DAT_DAY = DAT_START.replace(hour=0, minute=0, second=0, microsecond=0)


def check_clean_calendar(calendar):
    # Check the next few days are clean. Just a degenerate check for tests
    assert len(calendar.get_events(max_events=1, end=datetime.datetime.now() + datetime.timedelta(days=3))) == 0


def test_copy_event():
    """
    Tests at a basic level that events are copied from one calendar to the other
    """
    google_service = GoogleService()
    google_service.authorize(TEST_GOOGLE_USER_ID)
    google_calendar = google_service.get_calendar()

    microsoft_service = MicrosoftService()
    microsoft_service.authorize(TEST_MICROSOFT_USER_ID)
    microsoft_calendar = microsoft_service.get_calendar()

    check_clean_calendar(google_calendar)
    check_clean_calendar(microsoft_calendar)

    google_event_data = StubEvent(
        start=TIMEZONE.localize(TMW_START),
        end=TIMEZONE.localize(TMW_END),
        is_all_day=False,
    ).serialize(google_calendar.event_serializer)
    google_event_data = google_calendar.create_event(google_event_data, source_event_id=None)
    google_event = CreatedEvent.deserialize(google_event_data, google_calendar.event_serializer)

    microsoft_event_data = StubEvent(
        start=TIMEZONE.localize(DAT_START),
        end=TIMEZONE.localize(DAT_END),
        is_all_day=False,
    ).serialize(microsoft_calendar.event_serializer)
    microsoft_event_data = microsoft_calendar.create_event(microsoft_event_data, source_event_id=None)
    microsoft_event = CreatedEvent.deserialize(microsoft_event_data, microsoft_calendar.event_serializer)

    new_events, _, _ = synchronize([google_calendar, microsoft_calendar], max_days=3)

    google_calendar.delete_event(google_event.id, is_copy=False)
    for event in new_events[(1, 0)]:
        google_calendar.delete_event(event.id)

    microsoft_calendar.delete_event(microsoft_event.id)
    for event in new_events[(0, 1)]:
        microsoft_calendar.delete_event(event.id)

    assert len(new_events[(0, 1)] + new_events[(1, 0)]) == 2, f"Expected 2 sync'ed events. Got: {len(new_events[(0, 1)] + new_events[(1, 0)])}"
    assert StubEvent.from_(new_events[(0, 1)][0]) == StubEvent.from_(google_event)
    assert StubEvent.from_(new_events[(1, 0)][0]) == StubEvent.from_(microsoft_event)


def test_copy_recurring_event():
    """
    Tests that recurring events are copied over
    """
    google_service = GoogleService()
    google_service.authorize(TEST_GOOGLE_USER_ID)
    google_calendar = google_service.get_calendar()

    microsoft_service = MicrosoftService()
    microsoft_service.authorize(TEST_MICROSOFT_USER_ID)
    microsoft_calendar = microsoft_service.get_calendar()

    check_clean_calendar(google_calendar)
    check_clean_calendar(microsoft_calendar)

    google_event_data = StubEvent(
        start=TIMEZONE.localize(TMW_START),
        end=TIMEZONE.localize(TMW_END),
        is_all_day=False,
    ).serialize(google_calendar.event_serializer)
    google_event_data['recurrence'] = ['RRULE:FREQ=WEEKLY;COUNT=2']
    google_event_data = google_calendar.create_event(google_event_data, source_event_id=None)
    google_event = CreatedEvent.deserialize(google_event_data, google_calendar.event_serializer)

    microsoft_event_data = StubEvent(
        start=TIMEZONE.localize(DAT_START),
        end=TIMEZONE.localize(DAT_END),
        is_all_day=False,
    ).serialize(microsoft_calendar.event_serializer)
    dotw = DAT_START.strftime("%A").lower()
    microsoft_event_data['recurrence'] = {
        "pattern": {
            "type": "weekly",
            "interval": 1,  # Every week
            "daysOfWeek": [dotw],
            "firstDayOfWeek": dotw,
        },
        "range": {
            "type": "numbered",
            "numberOfOccurrences": 2,
            "startDate": DAT_START.strftime("%Y-%m-%d"),
        }
    }
    microsoft_event_data = microsoft_calendar.create_event(microsoft_event_data, source_event_id=None)
    microsoft_event = CreatedEvent.deserialize(microsoft_event_data, microsoft_calendar.event_serializer)

    new_events, _, _ = synchronize([google_calendar, microsoft_calendar], max_days=14) # include next two occurrences

    google_calendar.delete_event(google_event.id, is_copy=False)
    for event in new_events[(1, 0)]:
        google_calendar.delete_event(event.id)

    microsoft_calendar.delete_event(microsoft_event.id)
    for event in new_events[(0, 1)]:
        microsoft_calendar.delete_event(event.id)

    assert len(new_events[(0, 1)]) == 2, f"Expected 2 sync'ed events from Google to Microsoft. Got: {len(new_events[(0, 1)])}"
    assert len(new_events[(1, 0)]) == 2, f"Expected 2 sync'ed events from Microsoft to Google. Got: {len(new_events[(1, 0)])}"
    assert StubEvent.from_(new_events[(0, 1)][0]) == StubEvent.from_(google_event)
    assert StubEvent.from_(new_events[(1, 0)][0]) == StubEvent.from_(microsoft_event)


def test_copy_all_day_event():
    """
    Tests that all/multi-day events are copied successfully
    """
    google_service = GoogleService()
    google_service.authorize(TEST_GOOGLE_USER_ID)
    google_calendar = google_service.get_calendar()

    microsoft_service = MicrosoftService()
    microsoft_service.authorize(TEST_MICROSOFT_USER_ID)
    microsoft_calendar = microsoft_service.get_calendar()

    check_clean_calendar(google_calendar)
    check_clean_calendar(microsoft_calendar)

    google_event_data = StubEvent(
        start=TIMEZONE.localize(TMW_DAY),
        end=TIMEZONE.localize(TMW_DAY + datetime.timedelta(days=1)),
        is_all_day=True,
    ).serialize(google_calendar.event_serializer)
    google_event_data = google_calendar.create_event(google_event_data, source_event_id=None)
    google_event = CreatedEvent.deserialize(google_event_data, google_calendar.event_serializer)

    # NOTE: Microsoft needs date-aligned datetimes in order for all-day event
    # creation to succeed.
    microsoft_event_data = StubEvent(
        start=datetime.datetime(DAT_DAY.year, DAT_DAY.month, DAT_DAY.day, 0, 0, 0, tzinfo=pytz.utc),
        end=datetime.datetime(DAT_DAY.year, DAT_DAY.month, DAT_DAY.day + 1, 0, 0, 0, tzinfo=pytz.utc),
        is_all_day=True,
    ).serialize(microsoft_calendar.event_serializer)
    microsoft_event_data = microsoft_calendar.create_event(microsoft_event_data, source_event_id=None)
    microsoft_event = CreatedEvent.deserialize(microsoft_event_data, microsoft_calendar.event_serializer)

    new_events, _, _ = synchronize([google_calendar, microsoft_calendar], max_days=3)

    google_calendar.delete_event(google_event.id, is_copy=False)
    for event in new_events[(1, 0)]:
        google_calendar.delete_event(event.id)

    microsoft_calendar.delete_event(microsoft_event.id)
    for event in new_events[(0, 1)]:
        microsoft_calendar.delete_event(event.id)

    assert len(new_events[(0, 1)] + new_events[(1, 0)]) == 2, f"Expected 2 sync'ed events. Got: {len(new_events[(0, 1)] + new_events[(1, 0)])}"
    assert StubEvent.from_(new_events[(0, 1)][0]) == StubEvent.from_(google_event)
    assert StubEvent.from_(new_events[(1, 0)][0]) == StubEvent.from_(microsoft_event)


def test_update_edited_event():
    """
    Tests whether updates to edited events propagate to their sync'ed copies.
    """
    google_service = GoogleService()
    google_service.authorize(TEST_GOOGLE_USER_ID)
    google_calendar = google_service.get_calendar()

    microsoft_service = MicrosoftService()
    microsoft_service.authorize(TEST_MICROSOFT_USER_ID)
    microsoft_calendar = microsoft_service.get_calendar()

    check_clean_calendar(google_calendar)
    check_clean_calendar(microsoft_calendar)

    google_event_data = StubEvent(
        start=TIMEZONE.localize(TMW_START),
        end=TIMEZONE.localize(TMW_END),
        is_all_day=False,
    ).serialize(google_calendar.event_serializer)
    google_event_data = google_calendar.create_event(google_event_data, source_event_id=None)
    google_event = CreatedEvent.deserialize(google_event_data, google_calendar.event_serializer)

    microsoft_event_data = StubEvent(
        start=TIMEZONE.localize(DAT_START),
        end=TIMEZONE.localize(DAT_END),
        is_all_day=False,
    ).serialize(microsoft_calendar.event_serializer)
    microsoft_event_data = microsoft_calendar.create_event(microsoft_event_data, source_event_id=None)
    microsoft_event = CreatedEvent.deserialize(microsoft_event_data, microsoft_calendar.event_serializer)

    new_events1, _, _ = synchronize([google_calendar, microsoft_calendar], max_days=3)

    google_update_data = StubEvent(
        start=TIMEZONE.localize(DAT_START),
        end=TIMEZONE.localize(DAT_END),
        is_all_day=False,
    ).serialize(google_calendar.event_serializer)
    google_update_data = google_calendar.update_event(google_event.id, google_update_data, is_copy=False)
    google_update = CreatedEvent.deserialize(google_update_data, google_calendar.event_serializer)

    microsoft_update_data = StubEvent(
        start=TIMEZONE.localize(TMW_START),
        end=TIMEZONE.localize(TMW_END),
        is_all_day=False,
    ).serialize(microsoft_calendar.event_serializer)
    microsoft_update_data = microsoft_calendar.update_event(microsoft_event.id, microsoft_update_data)
    microsoft_update = CreatedEvent.deserialize(microsoft_update_data, microsoft_calendar.event_serializer)

    created2, updated2, deleted2 = synchronize([google_calendar, microsoft_calendar], max_days=3)

    google_calendar.delete_event(google_event.id, is_copy=False)
    for event in new_events1[(1, 0)] + created2[(1, 0)]:
        google_calendar.delete_event(event.id)

    microsoft_calendar.delete_event(microsoft_event.id)
    for event in new_events1[(0, 1)] + created2[(0, 1)]:
        microsoft_calendar.delete_event(event.id)

    # Check that update from Google to Microsoft calendar worked
    assert len(new_events1[(0, 1)]) == 1, f"Expected 1 sync'ed events. Got: {len(new_events1[(0, 1)])}"
    assert len(updated2[(0, 1)]) == 1, f"Expected 1 updated events. Got: {len(updated2[(0, 1)])}"
    assert len(created2[(0, 1)]) == 0, f"Expected 0 created events. Got: {len(created2[(0, 1)])}"
    assert StubEvent.from_(updated2[(0, 1)][0]) == StubEvent.from_(google_update)

    # Check that update from Microsoft to Google calendar worked
    assert len(new_events1[(1, 0)]) == 1, f"Expected 1 sync'ed events. Got: {len(new_events1[(0, 1)])}"
    assert len(updated2[(1, 0)]) == 1, f"Expected 1 updated events. Got: {len(updated2[(0, 1)])}"
    assert len(created2[(1, 0)]) == 0, f"Expected 0 created events. Got: {len(created2[(0, 1)])}"
    assert StubEvent.from_(updated2[(1, 0)][0]) == StubEvent.from_(microsoft_update)


def test_remove_deleted_event():
    google_service = GoogleService()
    google_service.authorize(TEST_GOOGLE_USER_ID)
    google_calendar = google_service.get_calendar()

    microsoft_service = MicrosoftService()
    microsoft_service.authorize(TEST_MICROSOFT_USER_ID)
    microsoft_calendar = microsoft_service.get_calendar()

    check_clean_calendar(google_calendar)
    check_clean_calendar(microsoft_calendar)

    google_event_data = StubEvent(
        start=TIMEZONE.localize(TMW_START),
        end=TIMEZONE.localize(TMW_END),
        is_all_day=False,
    ).serialize(google_calendar.event_serializer)
    google_event_data = google_calendar.create_event(google_event_data, source_event_id=None)
    google_event = CreatedEvent.deserialize(google_event_data, google_calendar.event_serializer)

    microsoft_event_data = StubEvent(
        start=TIMEZONE.localize(DAT_START),
        end=TIMEZONE.localize(DAT_END),
        is_all_day=False,
    ).serialize(microsoft_calendar.event_serializer)
    microsoft_event_data = microsoft_calendar.create_event(microsoft_event_data, source_event_id=None)
    microsoft_event = CreatedEvent.deserialize(microsoft_event_data, microsoft_calendar.event_serializer)

    new_events1, _, _ = synchronize([google_calendar, microsoft_calendar], max_days=3)

    google_calendar.delete_event(google_event.id, is_copy=False)
    microsoft_calendar.delete_event(microsoft_event.id)

    _, _, deleted_events2 = synchronize([google_calendar, microsoft_calendar], max_days=3)

    # Check that update from Google to Microsoft calendar worked
    assert len(new_events1[(0, 1)]) == 1, f"Expected 1 sync'ed events. Got: {len(new_events1[(0, 1)])}"
    assert len(new_events1[(1, 0)]) == 1, f"Expected 1 sync'ed events. Got: {len(new_events1[(0, 1)])}"
    assert len(deleted_events2[(1, 0)]) == 1, f"Expected 1 deleted events. Got: {len(deleted_events2[(1, 0)])}"
    check_clean_calendar(google_calendar)
    check_clean_calendar(microsoft_calendar)


def test_already_copied_event_microsoft():
    """
    Tests whether Microsoft events properly track (a) that it was createad by
    PotatoTime and (b) the source event
    """
    google_service = GoogleService()
    google_service.authorize(TEST_GOOGLE_USER_ID)
    google_calendar = google_service.get_calendar()

    microsoft_service = MicrosoftService()
    microsoft_service.authorize(TEST_MICROSOFT_USER_ID)
    microsoft_calendar = microsoft_service.get_calendar()

    check_clean_calendar(google_calendar)
    check_clean_calendar(microsoft_calendar)

    google_event_data = StubEvent(
        start=TIMEZONE.localize(TMW_START),
        end=TIMEZONE.localize(TMW_END),
        is_all_day=False,
    ).serialize(google_calendar.event_serializer)
    google_event_data = google_calendar.create_event(google_event_data, source_event_id=None)
    google_event = CreatedEvent.deserialize(google_event_data, google_calendar.event_serializer)

    new_events1, _, _ = synchronize([google_calendar, microsoft_calendar], max_days=3)
    new_events2, _, _ = synchronize([google_calendar, microsoft_calendar], max_days=3)

    google_calendar.delete_event(google_event.id, is_copy=False)
    for event in new_events1[(1, 0)] + new_events2[(1, 0)]:
        google_calendar.delete_event(event.id)

    for event in new_events1[(0, 1)] + new_events2[(0, 1)]:
        microsoft_calendar.delete_event(event.id)

    assert len(new_events1[(0, 1)] + new_events1[(1, 0)]) == 1, f"Expected 1 sync'ed events. Got: {len(new_events1[(0, 1)] + new_events1[(1, 0)])}"
    assert len(new_events2[(1, 0)]) == 0, f"Should not copy the copy PotatoTime made"
    assert len(new_events2[(0, 1)]) == 0, f"Should not sync the already-sync'ed event"
    assert StubEvent.from_(new_events1[(0, 1)][0]) == StubEvent.from_(google_event)


def test_already_copied_event_google():
    """
    Tests whether Google events properly track (a) that it was createad by
    PotatoTime and (b) the source event
    """
    google_service = GoogleService()
    google_service.authorize(TEST_GOOGLE_USER_ID)
    google_calendar = google_service.get_calendar()

    microsoft_service = MicrosoftService()
    microsoft_service.authorize(TEST_MICROSOFT_USER_ID)
    microsoft_calendar = microsoft_service.get_calendar()

    check_clean_calendar(google_calendar)
    check_clean_calendar(microsoft_calendar)

    microsoft_event_data = StubEvent(
        start=TIMEZONE.localize(DAT_START),
        end=TIMEZONE.localize(DAT_END),
        is_all_day=False,
    ).serialize(microsoft_calendar.event_serializer)
    microsoft_event_data = microsoft_calendar.create_event(microsoft_event_data, source_event_id=None)
    microsoft_event = CreatedEvent.deserialize(microsoft_event_data, microsoft_calendar.event_serializer)

    new_events1, _, _ = synchronize([microsoft_calendar, google_calendar], max_days=3)
    new_events2, _, _ = synchronize([microsoft_calendar, google_calendar], max_days=3)

    microsoft_calendar.delete_event(microsoft_event.id)
    for event in new_events1[(1, 0)] + new_events2[(1, 0)]:
        microsoft_calendar.delete_event(event.id)

    for event in new_events1[(0, 1)] + new_events2[(0, 1)]:
        google_calendar.delete_event(event.id)

    assert len(new_events1[(0, 1)] + new_events1[(1, 0)]) == 1, f"Expected 1 sync'ed events. Got: {len(new_events1[(0, 1)] + new_events1[(1, 0)])}"
    assert len(new_events2[(1, 0)]) == 0, f"Should not copy the copy PotatoTime made"
    assert len(new_events2[(0, 1)]) == 0, f"Should not sync the already-sync'ed event"
    assert StubEvent.from_(new_events1[(0, 1)][0]) == StubEvent.from_(microsoft_event)


# # TODO: automate setting up then declining an event
# def test_ignore_declined_google():
#     google_service = GoogleService()
#     google_service.authorize(TEST_GOOGLE_USER_ID)

#     microsoft_service = MicrosoftService()
#     microsoft_service.authorize(TEST_MICROSOFT_USER_ID)
#     microsoft_calendar = microsoft_service.get_calendar()

#     check_clean_calendar(microsoft_calendar)

#     new_events, _, _ = synchronize([google_calendar, microsoft_calendar], max_days=3)
#     assert len(new_events[(0, 1)] + new_events[(1, 0)]) == 0, f"Expected 0 sync'ed events. Got: {len(new_events)}"


# # # TODO: automate setting up then declining an event
# def test_ignore_declined_microsoft():
#     google_service = GoogleService()
#     google_service.authorize(TEST_GOOGLE_USER_ID)

#     microsoft_service = MicrosoftService()
#     microsoft_service.authorize(TEST_MICROSOFT_USER_ID)
#     microsoft_calendar = microsoft_service.get_calendar()

#     check_clean_calendar(google_calendar)

#     new_events, _, _ = synchronize([google_calendar, microsoft_calendar], max_days=3)
#     assert len(new_events[(0, 1)] + new_events[(1, 0)]) == 0, f"Expected 0 sync'ed events. Got: {len(new_events)}"



from potatotime.services import StubEvent, CreatedEvent
from potatotime.services.gcal import GoogleService
from potatotime.services.outlook import MicrosoftService
from potatotime.services.ical import AppleService
from utils import TIMEZONE, TEST_GOOGLE_USER_ID, TEST_MICROSOFT_USER_ID
import datetime
import pytz


def test_raw_google_calendar():
    google_service = GoogleService()
    google_service.authorize(TEST_GOOGLE_USER_ID)
    google_calendar = google_service.get_calendar()
    
    google_event_data = {
        'summary': 'Sample Event',
        'location': '800 Howard St., San Francisco, CA 94103',
        'description': 'A chance to hear more about Google\'s developer products.',
        'start': {
            'dateTime': '2024-08-01T09:00:00-07:00',
            'timeZone': 'US/Pacific',
        },
        'end': {
            'dateTime': '2024-08-01T17:00:00-07:00',
            'timeZone': 'US/Pacific',
        },
        'recurrence': [
            'RRULE:FREQ=WEEKLY;COUNT=10'
        ],
        'reminders': {
            'useDefault': False,
            'overrides': [
                {'method': 'email', 'minutes': 24 * 60},
                {'method': 'popup', 'minutes': 10},
            ],
        },
    }

    google_event = google_calendar.create_event(google_event_data, source_event_id='')
    google_update_data = {
        'summary': 'Updated Sample Event',
        'location': '500 Terry A Francois Blvd, San Francisco, CA 94158',
        'description': 'A chance to hear more about Google\'s updated developer products.',
        'start': {
            'dateTime': '2024-08-02T10:00:00-07:00',
            'timeZone': 'US/Pacific',
        },
        'end': {
            'dateTime': '2024-08-02T18:00:00-07:00',
            'timeZone': 'US/Pacific',
        },
    }
    google_calendar.update_event(google_event['id'], google_update_data)

    for event in google_calendar.get_events():
        start = event['start'].get('dateTime', event['start'].get('date'))
        print(start, event.get('summary'))

    google_calendar.delete_event(google_event['id'])


def test_google_calendar():
    google_service = GoogleService()
    google_service.authorize(TEST_GOOGLE_USER_ID)
    google_calendar = google_service.get_calendar()
    
    google_event_data = StubEvent(
        start=TIMEZONE.localize(datetime.datetime(2024, 8, 1, 10, 0, 0)),
        end=TIMEZONE.localize(datetime.datetime(2024, 8, 1, 11, 0, 0)),
        is_all_day=False,
    ).serialize(google_calendar.event_serializer)

    google_event_data = google_calendar.create_event(google_event_data, source_event_id='')
    google_event = CreatedEvent.deserialize(google_event_data, google_calendar.event_serializer)
    google_update_data = StubEvent(
        start=TIMEZONE.localize(datetime.datetime(2024, 8, 2, 10, 0, 0)),
        end=TIMEZONE.localize(datetime.datetime(2024, 8, 2, 11, 0, 0)),
        is_all_day=False,
    ).serialize(google_calendar.event_serializer)
    google_calendar.update_event(google_event.id, google_update_data)

    for event_data in google_calendar.get_events():
        event = CreatedEvent.deserialize(event_data, google_calendar.event_serializer)
        print(event.id, event.start, event.end)

    google_calendar.delete_event(google_event.id)


def test_raw_microsoft_calendar():
    microsoft_service = MicrosoftService()
    microsoft_service.authorize(TEST_MICROSOFT_USER_ID)
    microsoft_calendar = microsoft_service.get_calendar()
    
    microsoft_event_data = {
        "subject": "Discuss the new project",
        "body": {
            "contentType": "HTML",
            "content": "Let's discuss the new project on Thursday."
        },
        "start": {
            "dateTime": "2024-08-01T10:00:00",
            "timeZone": "Pacific Standard Time"
        },
        "end": {
            "dateTime": "2024-08-01T11:00:00",
            "timeZone": "Pacific Standard Time"
        },
        "location": {
            "displayName": "Conference Room"
        }
    }

    microsoft_event = microsoft_calendar.create_event(microsoft_event_data, source_event_id='')
    microsoft_update_data = {
        "subject": "Discuss the new project - Updated",
    }
    microsoft_calendar.update_event(microsoft_event['id'], microsoft_update_data)

    for event in microsoft_calendar.get_events():
        print(event['start']['dateTime'], event['subject'])

    microsoft_calendar.delete_event(microsoft_event['id'])


def test_microsoft_calendar():
    microsoft_service = MicrosoftService()
    microsoft_service.authorize(TEST_MICROSOFT_USER_ID)
    microsoft_calendar = microsoft_service.get_calendar()

    microsoft_event_data = StubEvent(
        start=TIMEZONE.localize(datetime.datetime(2024, 8, 1, 10, 0, 0)),
        end=TIMEZONE.localize(datetime.datetime(2024, 8, 1, 11, 0, 0)),
        is_all_day=False,
    ).serialize(microsoft_calendar.event_serializer)

    microsoft_event_data = microsoft_calendar.create_event(microsoft_event_data, source_event_id='')
    microsoft_event = CreatedEvent.deserialize(microsoft_event_data, microsoft_calendar.event_serializer)
    microsoft_update_data = StubEvent(
        start=TIMEZONE.localize(datetime.datetime(2024, 8, 2, 10, 0, 0)),
        end=TIMEZONE.localize(datetime.datetime(2024, 8, 2, 11, 0, 0)),
        is_all_day=False,
    ).serialize(microsoft_calendar.event_serializer)
    microsoft_calendar.update_event(microsoft_event.id, microsoft_update_data)

    for event_data in microsoft_calendar.get_events():
        event = CreatedEvent.deserialize(event_data, microsoft_calendar.event_serializer)
        print(event.id, event.start, event.end)

    microsoft_calendar.delete_event(microsoft_event.id)


def test_raw_apple_calendar():
    apple_service = AppleService()
    apple_service.authorize(None)  # user_id not currently used
    apple_calendar = apple_service.get_calendar()
    
    apple_event_data = {
        'start': datetime.datetime(2024, 8, 1, 10, 0, 0, tzinfo=pytz.utc),
        'end': datetime.datetime(2024, 8, 1, 11, 0, 0, tzinfo=pytz.utc),
        'summary': "New Event",
        'description': "This is a new event",
        'location': "Location"
    }
    apple_event = apple_calendar.create_event(apple_event_data)
    apple_update_data = {
        'start': datetime.datetime(2024, 8, 2, 10, 0, 0, tzinfo=pytz.utc),
        'end': datetime.datetime(2024, 8, 2, 11, 0, 0, tzinfo=pytz.utc),
        'summary': "Edited Event",
        'description': "This event has been edited",
        'location': "New Location"
    }
    apple_calendar.update_event(apple_event, apple_update_data)

    for event in apple_calendar.get_events():
        component = event.instance.vevent
        print(component.dtstart.value, component.summary.value)

    apple_calendar.delete_event(apple_event)


def test_apple_calendar():
    apple_service = AppleService()
    apple_service.authorize(None)  # user_id not currently used
    apple_calendar = apple_service.get_calendar()

    apple_event_data = StubEvent(
        start=datetime.datetime(2024, 8, 1, 10, 0, 0, tzinfo=pytz.utc),
        end=datetime.datetime(2024, 8, 1, 11, 0, 0, tzinfo=pytz.utc),
        is_all_day=False,
    ).serialize(apple_calendar.event_serializer)
    apple_event = apple_calendar.create_event(apple_event_data)
    apple_update_data = StubEvent(
        start=datetime.datetime(2024, 8, 2, 10, 0, 0, tzinfo=pytz.utc),
        end=datetime.datetime(2024, 8, 2, 11, 0, 0, tzinfo=pytz.utc),
        is_all_day=False,
    ).serialize(apple_calendar.event_serializer)
    apple_calendar.update_event(apple_event, apple_update_data)

    for event_data in apple_calendar.get_events():
        event = CreatedEvent.deserialize(event_data, apple_calendar.event_serializer)
        print(event.id, event.start, event.end)

    apple_calendar.delete_event(apple_event)



import datetime
import os.path
from urllib.error import HTTPError
from googleapiclient import errors
from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import build
from potatotime.services import ServiceInterface, CalendarInterface, EventSerializer, BaseEvent, POTATOTIME_EVENT_SUBJECT, POTATOTIME_EVENT_DESCRIPTION
from potatotime.services.auth import get_auth_code
from potatotime.storage import Storage, FileStorage
from typing import Optional, List, Dict, Union
import pytz
import json


class _GoogleEventSerializer(EventSerializer):
    def serialize(self, field_name: str, event: BaseEvent):
        if field_name in ('start', 'end'):
            field_value = getattr(event, field_name)
            if event.is_all_day:
                return field_name, {'date': field_value.strftime('%Y-%m-%d') }
            return field_name, {
                'dateTime': field_value.isoformat(),
                'timeZone': getattr(field_value.tzinfo, 'zone', field_value.tzname()) if field_value.tzinfo else 'UTC'
            }
        if field_name == 'is_all_day':
            return None, None
        raise NotImplementedError(f"Serializing {field_name} is not supported")
    
    def deserialize(self, field_name: str, event_data: dict):
        if field_name == 'id':
            return event_data.get('id')
        if field_name in ('start', 'end'):
            field_value = event_data[field_name]
            if 'dateTime' in field_value:
                return datetime.datetime.fromisoformat(field_value['dateTime'])
            elif 'date' in field_value:
                return pytz.utc.localize(datetime.datetime.fromisoformat(field_value['date']))
            raise NotImplementedError('Unsupported start and end time format')
        if field_name == 'url':
            return event_data.get('htmlLink')
        if field_name == 'source_event_id':
            return event_data.get('extendedProperties', {}).get('private', {}).get('potatotime')
        if field_name == 'declined':
            return any([attendee.get('self') and attendee['responseStatus'] == 'declined'
                        for attendee in event_data.get('attendees', [])])
        if field_name == 'is_all_day':
            return 'date' in event_data['start'] and 'date' in event_data['end']


class GoogleService(ServiceInterface):
    
    def __init__(self):
        # If modifying these SCOPES, delete the file goog.json.
        self.scopes = [
            "openid",
            "https://www.googleapis.com/auth/userinfo.profile",
            "https://www.googleapis.com/auth/userinfo.email",
            "https://www.googleapis.com/auth/calendar.events",
            "https://www.googleapis.com/auth/calendar.readonly",
        ]
        self.event_serializer = _GoogleEventSerializer()

    def authorize(self, user_id: str, storage: Storage=FileStorage(), interactive: bool=True):
        # TODO: This needs some major refactoring
        creds = None
        if storage.has_user_credentials(user_id):
            creds = Credentials.from_authorized_user_info(
                json.loads(storage.get_user_credentials(user_id)),
                self.scopes,
            )
        if not creds or not creds.valid:
            if creds and creds.expired and creds.refresh_token:
                # TODO: what if refresh fails? Get an exception. Maybe instead
                # gracefully default to interactive?
                creds.refresh(Request())
            elif interactive:
                # TODO: replace str 'google' with constant
                flow = InstalledAppFlow.from_client_config(
                    json.loads(storage.get_client_credentials('google')),
                    self.scopes
                )
                flow.redirect_uri = 'http://localhost:5173/'
                auth_url, _ = flow.authorization_url(access_type='offline', prompt='consent')
                auth_code = get_auth_code(auth_url, port=5173)
                flow.fetch_token(code=auth_code)
                creds = flow.credentials
            try:
                storage.save_user_credentials(user_id, creds.to_json())
            except Exception as e:
                print(f"Failed to save updated credentials: {e}")

            if not creds:
                raise Exception('No credentials found, or credentials are expired.')
        self.service = build('calendar', 'v3', credentials=creds)

    def list_calendars(self) -> List[Dict]:
        try:
            calendar_list = self.service.calendarList().list().execute()
            return calendar_list.get('items', [])
        except HTTPError as error:
            print(f'An error occurred: {error}')
            return []
    
    def get_calendar(self, calendar_id: Optional[str]=None):
        calendars = self.list_calendars()
        for calendar in calendars:
            if calendar['id'] == calendar_id or calendar_id is None:
                return GoogleCalendar(self.service, calendar_id)
        raise ValueError(f'Invalid calendar_id: {calendar_id}')


class GoogleCalendar(CalendarInterface):

    def __init__(self, service, calendar_id):
        self.service = service
        self.calendar_id = calendar_id
        self.event_serializer = _GoogleEventSerializer()
    
    def get_events(
        self,
        start: Optional[datetime.datetime]=None,
        end: Optional[datetime.datetime]=None,
        max_events: int=1000,
        results_per_page: int=100,
    ):
        if not start:
            start = datetime.datetime.utcnow()
        if not end:
            end = start + datetime.timedelta(days=30)
        
        events = []
        page_token = None

        while True:
            events_result = self.service.events().list(
                calendarId='primary',
                timeMin=start.isoformat() + 'Z',
                timeMax=end.isoformat() + 'Z',
                maxResults=min(results_per_page, max_events - len(events)),  # Ensure we do not exceed max_events
                singleEvents=True,
                orderBy='startTime',
                pageToken=page_token
            ).execute()

            events.extend(events_result.get('items', []))
            
            # Check if we have reached the maximum number of events
            if len(events) >= max_events:
                events = events[:max_events]
                break
            
            # Get the next page token, if there is one
            page_token = events_result.get('nextPageToken')
            if not page_token:
                break

        return events

    def create_event(self, event_data: dict, source_event_id: Optional[str]):
        if source_event_id is not None:  # NOTE: Should only be None during testing
            event_data['extendedProperties'] = {'private': {'potatotime': source_event_id}}
        event_data['summary'] = POTATOTIME_EVENT_SUBJECT
        event_data['description'] = POTATOTIME_EVENT_DESCRIPTION
        event_data['colorId'] = '8'  # Light gray color
        event = self.service.events().insert(calendarId='primary', body=event_data).execute()
        print(f'Event created: {event.get("htmlLink")}')
        return event

    def update_event(self, event_id, update_data, is_copy: bool=True):
        event = self.service.events().get(calendarId='primary', eventId=event_id).execute()
        if is_copy:  # NOTE: Should only be False during testing
            assert 'potatotime' in event.get('extendedProperties', {}).get('private', {})
        event.update(update_data)
        updated_event = self.service.events().update(calendarId='primary', eventId=event_id, body=event).execute()
        print(f'Event updated: {updated_event.get("htmlLink")}')
        return updated_event

    def delete_event(self, event_or_event_id: Union[str, dict], is_copy: bool=True):
        # TODO: Only this gcal implementation supports raw event_data dict. Update interface
        if isinstance(event_or_event_id, dict):
            event = event_or_event_id
            event_id = event['id']
        else:
            event_id = event_or_event_id
            event = self.service.events().get(calendarId='primary', eventId=event_id).execute()
        if is_copy:  # NOTE: Should only be False during testing
            assert 'potatotime' in event.get('extendedProperties', {}).get('private', {})
        try:
            self.service.events().delete(calendarId='primary', eventId=event_id).execute()
            print(f'Event "{event_id}" deleted.')
        except errors.HttpError as error:
            print(f'An error occurred: {error}')


if __name__ == '__main__':
    service = GoogleService()
    service.authorize('default_google')  # TODO: use constant for user_id


import os
from typing import Dict
import json
from abc import ABC, abstractmethod


class Storage(ABC):
    @abstractmethod
    def has_user_credentials(self, user_id: str) -> bool:
        pass

    @abstractmethod
    def get_user_credentials(self, user_id: str) -> Dict:
        pass

    @abstractmethod
    def save_user_credentials(self, user_id: str, credentials: str):
        pass

    @abstractmethod
    def get_client_credentials(self, client_id: str):
        pass


class FileStorage(Storage):
    TEMPLATE_USER = "potatotime_user_{user_id}.json"
    TEMPLATE_CLIENT = "potatotime_client_{client_id}.json"

    def has_user_credentials(self, user_id: str) -> bool:
        return os.path.exists(self.TEMPLATE_USER.format(user_id=user_id))

    def get_user_credentials(self, user_id: str) -> Dict:
        if self.has_user_credentials(user_id):
            with open(self.TEMPLATE_USER.format(user_id=user_id)) as f:
                return f.read()

    def save_user_credentials(self, user_id: str, credentials: str):
        # TODO: json.dumps here, to be consistent?
        with open(self.TEMPLATE_USER.format(user_id=user_id), 'w') as f:
            f.write(credentials)

    def get_client_credentials(self, client_id: str):
        with open(self.TEMPLATE_CLIENT.format(client_id=client_id)) as f:
            return f.read()


class EnvStorage(Storage):
    """
    Environment variables for holding credentials. Note that writes are not
    supported.
    """
    TEMPLATE_USER = "POTATOTIME_USER_{user_id}"
    TEMPLATE_CLIENT = "POTATOTIME_CLIENT_{client_id}"

    def has_user_credentials(self, user_id: str) -> bool:
        return self.TEMPLATE_USER.format(user_id=user_id) in os.environ

    def get_user_credentials(self, user_id: str) -> Dict:
        return os.environ.get(self.TEMPLATE_USER.format(user_id=user_id), '{}')

    def save_user_credentials(self, user_id: str, credentials: str):
        raise NotImplementedError('Writes are not implemented for EnvStorage')

    def get_client_credentials(self, client_id: str):
        return os.environ.get(self.TEMPLATE_CLIENT.format(client_id=client_id), '{}')



from setuptools import setup, find_packages

setup(
    name='potatotime',
    version='0.0.5',
    packages=find_packages(),
    install_requires=[
        # List your dependencies here
        # e.g., 'numpy', 'requests'
    ],
    entry_points={
        'console_scripts': [
            # Define command-line scripts if needed
            # 'your_command=your_module:main_function'
        ],
    },
)